/*******************************************************************************
 * Licensed Materials - Property of IBM
 * CompareBuilds.java
 * (c) Copyright IBM Corporation 2012. 
 * U.S. Government Users Restricted Rights:  Use, duplication or disclosure 
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

package net.jazz.plainjava.check.approvals;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;

import com.ibm.team.build.client.ITeamBuildClient;
import com.ibm.team.build.common.ScmConstants;
import com.ibm.team.build.common.model.BuildState;
import com.ibm.team.build.common.model.IBuildDefinition;
import com.ibm.team.build.common.model.IBuildResult;
import com.ibm.team.build.common.model.IBuildResultContribution;
import com.ibm.team.build.common.model.query.IBaseBuildResultQueryModel.IBuildResultQueryModel;
import com.ibm.team.links.common.ILink;
import com.ibm.team.links.common.ILinkHandle;
import com.ibm.team.repository.client.IItemManager;
import com.ibm.team.repository.client.ILoginHandler2;
import com.ibm.team.repository.client.ILoginInfo2;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.client.TeamPlatform;
import com.ibm.team.repository.client.login.UsernameAndPasswordLoginInfo;
import com.ibm.team.repository.common.IItemHandle;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.repository.common.query.IItemQuery;
import com.ibm.team.repository.common.query.IItemQueryPage;
import com.ibm.team.repository.common.query.ast.IPredicate;
import com.ibm.team.repository.common.service.IQueryService;
import com.ibm.team.scm.client.IWorkspaceManager;
import com.ibm.team.scm.common.IBaselineSetHandle;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.dto.IChangeHistorySyncReport;
import com.ibm.team.scm.common.dto.IChangeSetLinkSummary;
import com.ibm.team.workitem.common.model.IApproval;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemHandle;
import com.ibm.team.workitem.common.model.WorkItemApprovals;


/**
 * <p>
 * In this example, two builds, either the two latest or two provided as arguments to the program,
 * are examined to discover whether the change sets that are delivered have work items with approvals
 * on them. The work items must have approvals and be in an approved state or a warning is printed to
 * standard out.
 * </p>
 */
public class CompareBuilds {
	public static void main(final String[] args) throws TeamRepositoryException {
		if (args.length < 4) {
			System.out.println("Please provide repository URI, username, password and build ID as first four arguments to this program.");
			System.exit(1);
		}
		
		String repoUri = args[0];
		final String userId = args[1];
		final String password = args[2];
		
		TeamPlatform.startup();
		
		try {
			// Login to the repository using the provided credentials
			ITeamRepository repo = TeamPlatform.getTeamRepositoryService().getTeamRepository(repoUri);
			repo.registerLoginHandler(new ILoginHandler2() {
				@Override
				public ILoginInfo2 challenge(ITeamRepository arg0) {
					return new UsernameAndPasswordLoginInfo(userId, password);
				}
			});
			repo.login(new SysoutProgressMonitor());
			
			// Gather the necessary clients and managers necessary from the repository
			// In this case we need the build client, source control workspace manager and the item manager
			ITeamBuildClient buildClient = (ITeamBuildClient) repo.getClientLibrary(ITeamBuildClient.class);
			IWorkspaceManager workspaceManager = (IWorkspaceManager)repo.getClientLibrary(IWorkspaceManager.class);
			IItemManager itemManager = repo.itemManager();
			
			
			// Find the build definition on the server
			String buildId = args[3];
			
			IBuildDefinition definition = buildClient.getBuildDefinition(buildId, new SysoutProgressMonitor());
			
			if (definition == null) {
				System.out.println(buildId+": build definition not found.");
				System.exit(1);
			}
			
			// Construct a query to find completed builds that aren't personal builds.
	        IBuildResultQueryModel buildResultQueryModel = IBuildResultQueryModel.ROOT;
	        IItemQuery query = IItemQuery.FACTORY.newInstance(buildResultQueryModel);
	        
	        IPredicate predicate = (buildResultQueryModel.buildDefinition()._eq(query.newItemHandleArg()))._and(
	    	        buildResultQueryModel.personalBuild()._isFalse())._and(
	    	    	        buildResultQueryModel.buildState()._eq(BuildState.COMPLETED.name()));
	        
	        // If build result ID's are provided then restrict to only those two results as well.
	        // Otherwise we will work with the two latest build results.
	        if (args.length >= 6) {
	        	predicate = predicate._and(buildResultQueryModel.label()._eq(args[4]).
	        			_or(buildResultQueryModel.label()._eq(args[5])));
	        }
	        
	        // Add predicate filter to the query
	        query.filter(predicate);
	        
	        // Order by the latest build first
	        query.orderByDsc(buildResultQueryModel.buildStartTime());

	        // We only need the top two results
	        query.setResultLimit(2);
	        
	        // Execute the query
	        IItemQueryPage queryPage = buildClient.queryItems(query, new Object[] { definition.getItemId() },
	                    IQueryService.ITEM_QUERY_MAX_PAGE_SIZE, new SysoutProgressMonitor());

	        // No results found. Maybe there are no complete, non-personal builds? Maybe the build ID's are wrong or cleaned up?
	        if (queryPage.getSize() == 0) {
	            System.out.println("Could not find the two builds. Make sure that there are two build results that are complete and not personal. Also, check the provided build ID's (if any).");
	        } else {
	        	// Fetch the complete build result items from the first two build result handles in the query results.
	        	@SuppressWarnings("unchecked")
				List<IBuildResult> buildResults = itemManager.fetchCompleteItems(queryPage.getItemHandles().subList(0, 2), IItemManager.DEFAULT, new SysoutProgressMonitor());
	   
	            IBuildResult latest = buildResults.get(0);
	            IBuildResult secondLatest = buildResults.get(1);
	            
	            System.out.println("Approvals for changesets delivered between "+latest.getLabel()+" and "+secondLatest.getLabel()+":\n");
	            
	            // Source control snapshots come from a special build result contribution on each build result
	            IBuildResultContribution[] snapshotContribution1 = buildClient.getBuildResultContributions(latest,
	                    new String[] { ScmConstants.EXTENDED_DATA_TYPE_ID_BUILD_SNAPSHOT }, new SysoutProgressMonitor());
	           
	            IBuildResultContribution[] snapshotContribution2 = buildClient.getBuildResultContributions(secondLatest,
	                    new String[] { ScmConstants.EXTENDED_DATA_TYPE_ID_BUILD_SNAPSHOT }, new SysoutProgressMonitor());

	            if (snapshotContribution1.length == 0 || snapshotContribution2.length == 0) {
	            	System.out.println("One of the selected builds "+latest.getLabel()+" "+secondLatest.getLabel()+" did not contain a source control snapshot to use in the comparison.");
	            } else {
	                IBaselineSetHandle baselineSetHandle1 = (IBaselineSetHandle) snapshotContribution1[0].getExtendedContribution();
	                IBaselineSetHandle baselineSetHandle2 = (IBaselineSetHandle) snapshotContribution2[0].getExtendedContribution();

	                // Create a sync report comparing the two snapshots so that the new change sets can be found
	                IChangeHistorySyncReport syncReport = workspaceManager.compareBaselineSets(baselineSetHandle1, baselineSetHandle2, null, new SysoutProgressMonitor());
	               
	                // Fetch the link summaries of the change sets listed in the sync report so that the linked workitems (if any) can
	                //  be found.
	                @SuppressWarnings("unchecked")
					List<IChangeSetHandle> changeSets = syncReport.outgoingChangeSets();
	                if (changeSets.isEmpty()) {
	                	System.out.println("No change sets delivered between the two builds.");
	                	System.exit(0);
	                }
	                Collection<IChangeSetLinkSummary> linkSummaries = workspaceManager.getChangeSetLinkSummary(changeSets, new SysoutProgressMonitor());
	                
	                // Collect all of the link handles for the change sets so that we can do one fetch of all of the links
	                List<ILinkHandle> linkHandles = new ArrayList<ILinkHandle>();
	                for (IChangeSetLinkSummary linkSummary: linkSummaries) {
	                	linkHandles.addAll(linkSummary.getLinks());
	                }
	                
	                // Some of the change sets could be attached to the same work item. Although work item handles
	                //  do have a unique ID those objects don't implement equals so we use a map of a work item handle
	                //  UUID to a work item handle with that UUID.
	                HashMap<UUID, IWorkItemHandle> workItemHandles = new HashMap<UUID, IWorkItemHandle>();
                	for (Object o: itemManager.fetchCompleteItems(linkHandles, IItemManager.DEFAULT, new SysoutProgressMonitor())) {
                		ILink link = (ILink) o;
                		
                		// Any links that aren't item references are not work item links so discard them now
                		if (link.getTargetRef().isItemReference()) {
                			IItemHandle itemHandle = (IItemHandle) link.getTargetRef().resolve();
                			
                			if (itemHandle instanceof IWorkItemHandle) {
                				workItemHandles.put(itemHandle.getItemId(), (IWorkItemHandle)itemHandle);
                			}
                		}
                	}
                	
                	// Fetch all of the work items so that we can inspect the approvals
                	for (Object item: itemManager.fetchCompleteItems(new ArrayList<IItemHandle>(workItemHandles.values()), IItemManager.DEFAULT, new SysoutProgressMonitor())) {
                		IWorkItem workItem = (IWorkItem)item;
                		
                		System.out.println("[" + workItem.getId() + "] " + workItem.getHTMLSummary().getPlainText());
                		
                		for (IApproval approval: workItem.getApprovals().getContents()) {
                			if (!WorkItemApprovals.APPROVED_STATE.getIdentifier().equals(approval.getStateIdentifier())){
                				System.out.println("\tNOT APPROVED:" + approval.getDescriptor().getName());
                			}
                		}
                		
                		if (workItem.getApprovals().getContents().size() == 0) {
                			System.out.println("\tNO APPROVALS");
                		}
                	}
	            } 
	        }
		} finally {
			TeamPlatform.shutdown();
		}
	}
}
